#ifndef FRAME_CPP__Common__CompressionZeroSuppress_hh
#define	FRAME_CPP__Common__CompressionZeroSuppress_hh

#include "general/types.hh"
#include "general/SharedArray.hh"

#include "framecpp/Common/CompressionDifferential.hh"
#include "framecpp/Common/CompressionException.hh"

namespace FrameCPP
{
  //=====================================================================
  // Exceptions
  //=====================================================================

  class FrZCompRangeError
    : public FrUncompressable
  {
  public:
    inline FrZCompRangeError( INT_2S Value )
      : FrUncompressable( format_msg( Value ) )
    {
    };

  private:
    inline std::string
    format_msg( INT_2S Value )
    {
      std::ostringstream	oss;

      oss << "Buffer cannot be compressed using FrVectZCompShort because value of "
	  << Value << " encountered";
      return oss.str( );
    }
  };

  class FrZCompIRangeError
    : public FrUncompressable
  {
  public:
    inline FrZCompIRangeError( INT_4S Value )
      : FrUncompressable( format_msg( Value ) )
    {
    };
  private:
    inline std::string
    format_msg( INT_4S Value )
    {
      std::ostringstream	oss;

      oss << "Buffer cannot be compressed using FrVectZCompInt because value of "
	  << Value << " encountered";
      return oss.str( );
    }
  };

  namespace Compression
  {
    //-------------------------------------------------------------------
    /// \brief Routines related to the gzip compression algorith
    //-------------------------------------------------------------------
    namespace ZeroSuppress
    {
      typedef General::SharedArray< CHAR_U > data_type;
      typedef data_type::element_type data_element_type;

      //-----------------------------------------------------------------
      /// \brief Compress the data with ZeroSuppress.
      ///
      /// \param[in] DataIn
      ///     Pointer to the data to be compressed
      /// \param[in] NBytesIn
      ///     The number of bytes to be compressed.
      ///     Upon return, the size of the compressed buffer is returned.
      /// \param[out] DataOut
      ///     Pointer to the data to be compressed
      /// \param[out] NBytesOut
      ///     The number of bytes to be compressed.
      ///     Upon return, the size of the compressed buffer is returned.
      /// \todo Correct this documentation
      //-----------------------------------------------------------------
      template < int > void
      Compress( const data_element_type* DataIn, INT_8U NBytesIn,
		    General::AutoArray< data_element_type >& DataOut, INT_8U& NBytesOut );

      template < int > void
      Compress( const data_element_type* DataIn, INT_8U NBytesIn,
		data_type& DataOut, INT_8U& NBytesOut );

      template < int > void
      Expand( const data_element_type* DataIn, INT_8U DataInBytes,
	      bool ByteSwap,
	      General::AutoArray< data_element_type >& DataOut, INT_8U NDataOut,
	      INT_8U& NBytesOut );

      template < int > void
      Expand( const data_element_type* DataIn, INT_8U DataInBytes,
	      bool ByteSwap,
	      data_type& DataOut, INT_8U NDataOut,
	      INT_8U& NBytesOut );

      template < int > void
      DiffCompress( const data_element_type* DataIn, INT_8U NBytesIn, INT_4U DataType,
		    General::AutoArray< data_element_type >& DataOut, INT_8U& NBytesOut );

      template < int > void
      DiffCompress( const data_element_type* DataIn, INT_8U NBytesIn, INT_4U DataType,
		    data_type& DataOut, INT_8U& NBytesOut );

      template < int Size > void
      DiffExpand( const data_element_type* DataIn, INT_8U DataInBytes,
		  bool ByteSwap, INT_4U DataType,
		  General::AutoArray< data_element_type >& DataOut, INT_8U NDataOut,
		  INT_8U& NBytesOut );

      template < int Size > void
      DiffExpand( const data_element_type* DataIn, INT_8U DataInBytes,
		  bool ByteSwap, INT_4U DataType,
		  data_type& DataOut, INT_8U NDataOut,
		  INT_8U& NBytesOut );

    } // namespace - ZeroSuppress

  } // namespace -  Compression
  
} // namespace - FrameCPP


#endif /* FRAME_CPP__Common__CompressionZeroSuppress_hh */
